
! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

MODULE DRAINE_DUST

USE constantes

IMPLICIT NONE

!  Draine's files give log grids for Q_abs, Q_scat, g. Change the following parameters if you change file.

   INTEGER, PARAMETER, PUBLIC                           :: nptmax=81          ! Number of radius pts max
   INTEGER, PARAMETER, PUBLIC                           :: npwmax=241         ! Number of wlgth pts max

!  Grains data base. Read in READ_DRAINE_FILES

   INTEGER, PUBLIC                                      :: np_r                ! Number of radius pts read
   INTEGER, PUBLIC                                      :: np_w                ! Number of wlgth pts read

   REAL (KIND=dp), PUBLIC                               :: r_min, r_max   
   REAL (KIND=dp), DIMENSION (nptmax), PUBLIC           :: radgr
   REAL (KIND=dp), DIMENSION (npwmax), PUBLIC           :: wavgrd

!  Optical properties

   REAL (KIND=dp), DIMENSION (nptmax,npwmax), PUBLIC    :: Qabs_Gra, Qsca_Gra, g_Gra
   REAL (KIND=dp), DIMENSION (nptmax,npwmax), PUBLIC    :: Qabs_Sil, Qsca_Sil, g_Sil
   REAL (KIND=dp), PUBLIC, DIMENSION (:,:), ALLOCATABLE :: Qabs

CONTAINS

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!     READING DATA FILES GIVEN BY DRAINE FOR GRAPHITE AND SILICATES
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     SUBROUTINE READ_DRAINE_FILES
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

       USE constantes
       USE inoutput

       IMPLICIT NONE

       INTEGER                   :: i, j
       REAL (KIND=dp)            :: toto


!---------------------------------------------------------------------
!             READING  GRAPHITE and SILICATE IN FILE:  gr_GS_81
!---------------------------------------------------------------------

       write(iscrn,'(/,"  Lecture des donnees de Draine",/)')

       fichier = TRIM(data_dir)//'gr_GS_81'
       OPEN (2, file=fichier, status='old')

       READ (2,'(3/,i4,2e10.3)') np_r, r_min, r_max
       READ (2,*)                np_w

       if (np_r /= nptmax) then
          print *, " !!! WRONG number of radii !!!"
          stop
       endif
       if (np_w /= npwmax) then
          print *, " !!! WRONG number of wavelengths !!!"
          stop
       endif

       DO i = 1,np_r
          READ (2,*) radgr(i)
          READ( 2,*)
          DO j = np_w,1,-1
             READ (2,*) wavgrd(j), Qabs_Gra(i,j), Qsca_Gra(i,j), g_Gra(i,j) &
                      ,   toto, Qabs_Sil(i,j), Qsca_Sil(i,j), g_Sil(i,j)
             if (toto /= wavgrd(j)) then
               print *, " pb lecture!"
             endif
          END DO
       END DO
       CLOSE(2)

       radgr = radgr * 1.0e-4_dp                               ! Convert from micrometer to cm
       wavgrd = wavgrd * 1.0e4_dp                              ! Convert from micrometer to Angstrom

     END SUBROUTINE READ_DRAINE_FILES

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    REAL(KIND=dp)  FUNCTION Q_inter(Qx,x,a,i0)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

      IMPLICIT NONE

      REAL(KIND=dp), DIMENSION(:), INTENT(in)               :: Qx
      REAL(KIND=dp), DIMENSION(:), INTENT(in)               :: x
      REAL(KIND=dp), INTENT(in)                             :: a
      INTEGER, INTENT(in)                                   :: i0

      Q_inter = Qx(i0) + (Qx(i0+1)-Qx(i0)) * ((x(i0+1)-a) / (x(i0+1)-x(i0)))

    END FUNCTION Q_inter

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   subroutine MIXGRSI (wlmin, wlmax, jmin, jmax)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   USE constantes
   USE auxiliaries
   USE PROFIL

   IMPLICIT NONE

    REAL (KIND=dp), INTENT (IN)        :: wlmin, wlmax
    INTEGER, INTENT (OUT)              :: jmin, jmax

    REAL (KIND=dp)                     :: amix = 0.3_dp
    REAL (KIND=dp)                     :: Qg1, Qg2, Qs1, Qs2
    INTEGER                            :: i0, i, j, mm

! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

! Find range in wavelength grid

   jmin = 0
   jmax = 0
   do i=1,np_w-1
     if (wavgrd(i) <= wlmin .AND. wavgrd(i+1) > wlmin) then
       jmin = i
     endif
     if (wavgrd(i) <= wlmax .AND. wavgrd(i+1) > wlmax) then
       jmax = i
       exit
     endif
   end do
   if (jmax == 0) then
     jmax = np_w
   endif

   allocate (Qabs(jmax-jmin+2,npg))
   Qabs = 0.0_dp

   do mm =1,npg
     if (adust(mm) < radgr(1) .OR. adust(mm) > radgr(np_r)) then
       print *, " Radius out of range:", radgr(1), " -", radgr(np_r)
       stop
     endif

     i0 = 1
     do
       if (radgr(i0+1) > adust(mm) .AND. radgr(i0) <= adust(mm)) then
         exit
       endif
       i0 = i0 + 1
     enddo

     Qg1 = Q_inter(Qabs_Gra(:,jmin),radgr,adust(mm),i0)
     Qg2 = Q_inter(Qabs_Gra(:,jmin+1),radgr,adust(mm),i0)
     Qs1 = Q_inter(Qabs_Sil(:,jmin),radgr,adust(mm),i0)
     Qs2 = Q_inter(Qabs_Sil(:,jmin+1),radgr,adust(mm),i0)
     Qabs(1,mm) = (1.0_dp-amix) * (Qg1 + (Qg2-Qg1) * (wlmin-wavgrd(jmin)) / (wavgrd(jmin+1)-wavgrd(jmin))) &
                +         amix  * (Qs1 + (Qs2-Qs1) * (wlmin-wavgrd(jmin)) / (wavgrd(jmin+1)-wavgrd(jmin)))

     do j=jmin+1,jmax
       Qg1 = Q_inter(Qabs_Gra(:,j),radgr,adust(mm),i0)
       Qs1 = Q_inter(Qabs_Sil(:,j),radgr,adust(mm),i0)
       Qabs(j-jmin+1,mm) = (1.0_dp-amix) * Qg1 + amix * Qs1
     end do

     Qg1 = Q_inter(Qabs_Gra(:,jmax),radgr,adust(mm),i0)
     Qg2 = Q_inter(Qabs_Gra(:,jmax+1),radgr,adust(mm),i0)
     Qs1 = Q_inter(Qabs_Sil(:,jmax),radgr,adust(mm),i0)
     Qs2 = Q_inter(Qabs_Sil(:,jmax+1),radgr,adust(mm),i0)
     Qabs(jmax-jmin+2,mm) = (1.0_dp-amix) * (Qg1 + (Qg2-Qg1) * (wlmax-wavgrd(jmax)) / (wavgrd(jmax+1)-wavgrd(jmax))) &
                          +         amix  * (Qs1 + (Qs2-Qs1) * (wlmax-wavgrd(jmax)) / (wavgrd(jmax+1)-wavgrd(jmax)))

     Qabs(:,mm) = Qabs(:,mm) * xpi * adust(mm)**2 / x_nc(mm)

   end do

!  write (73,'("# a_gr",1p,12e12.4)') (adust(i), i=1,npg)
!  write (73,'(1p,26e12.4)') wlmin, (Qabs(1,i), i=1,npg)
!  do j=2,jmax-jmin+1
!    write (73,'(1p,26e12.4)') wavgrd(j+jmin-1), (Qabs(j,i), i=1,npg)
!  end do
!  write (73,'(1p,26e12.4)') wlmax, (Qabs(jmax-jmin+2,i), i=1,npg)

   end subroutine MIXGRSI

  END MODULE DRAINE_DUST

