
! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

!*****************************************************************************************
! Radiation Field manipulations tools
!*****************************************************************************************
MODULE RF_TOOLS

USE constantes

IMPLICIT NONE

! RF_TYPE is used for all wavelength dependant quantities.
!         indexes go from lower to upper (the user does have to care if lower = 0 or not)
!         Values are in member Val. They must be ordered in increasing order of WaveLength
!         Ord is used to retrieve the Wave Length value from object rf_wl
!         rf_wl%Ord(i) is the rank at which rf_wl%val(i) has been entered in rf_wl
!         A specific vector (wl_pointer) is used to retrieve i from a given iord
!         Thus rf_wl may been unevenly spaced, and values entered in ony order

! A special set of subroutines and functions is provided to manipulate RF_TYPE objects
!
!   SUBROUTINE rf_create (rf_xx, ll, uu)
!              creates object rf_xx, and allocates Val and Ord from ll to uu
!   SUBROUTINE rf_free (rf_xx)
!              deallocates rf_xx%Val and rf_xx%Ord
!   SUBROUTINE rf_extand (rf_xx, x, add_1, iloc, iord)
!              add values x in rf_xx%Val. Val must be orderered in ascending order
!              x is inserted at its proper rank and the routine returns its location (iloc)
!              and its ordering number (iord). Note that iloc may vary as new values are added
!              but iord never changes.
!              If x is allready in rf_xx, add_1 returns with value 0. If not, add_1 = 1
!              rf_extand was created and is currently used only to add new wavelengths in
!              rf_wl
!   SUBROUTINE rf_locate (rf_xx, x, j)
!              returns index j such that rf_xx%Val(i) <= x < rf_xx%Val(i+1)
!              if x is outside rf_xx%Val, returns either lower or upper as appropriate
!              rf_xx%Val must be ordered in ascending order.
!   SUBROUTINE Shell_sort (rf_xxin, rf_xxout)
!              Sorts rf_object rf_xxin in ascending order
!   SUBROUTINE rf_revers (rindex)
!              Builds vector wl_pointer used to retrieve location of a given wavelength
!              from its ordering number.
!   REAL(KIND=dp) FUNCTION  rf_integration (xx, ffx)
!              Integrates ffx over xx using a trapezium rule.
!              xx and ffx are both rf_objects and must have the same size
!   REAL(KIND=dp) FUNCTION  sect_int (sect, rf_fn, rindex, wl_thresh)
!              Integrates the product sect * rf_fn over wavelength up to wl_thresh.
!              sect may have fewer points than rf_wl, but all wl must be in rf_wl (usually
!              added via rf_extand). rf_fn must have the same number of points as rf_wl
!              (usually, Radiation field intensity if sect is a cross section).
!              sect is interpolated linearly between two consecutive points, and the
!              integral between those points is performed by rf_integration.
!

TYPE RF_TYPE !------------------------------------------------------------------------------
  INTEGER                                          :: lower             !  Lower index     !
  INTEGER                                          :: upper             !  Upper index     !
  REAL (KIND=dp), DIMENSION (:), POINTER           :: Val               !  Values....      !
! REAL (KIND=dp), DIMENSION (:), POINTER           :: Val => NULL()     !  Values....      !
  INTEGER, DIMENSION (:), POINTER                  :: Ord               !  Order           !
END TYPE RF_TYPE !--------------------------------------------------------------------------

Type PHOTO_ESP !----------------------------------------------------------------------------
! used for all wavelength dependant quantities.                                            !
! This is a table containing photodestruction reactions parameters on which we based new   !
! calculation of their corresponding photodestruction rates. This new calculations         !
! integrating the FUV/UV radiation field at each point of the code used update             ! 
! photodestruction cross sections listed in files 'data/section/*.dat'.                    !
! With the PHOTO_ESP type, we succeed to stock these sections and other useful             ! 
! photo-reaction parameters. In this new TYPE, we also defined a variable as a RF_TYPE     !
! variable.                                                                                !
                                                                                           ! 
  CHARACTER (LEN=5)        :: nom                                                          !
  CHARACTER (LEN=8)        :: r1, r2, r3        ! Reagents in the photo-reaction           !
  CHARACTER (LEN=8)        :: p1, p2, p3, p4    ! Products from the photo-reaction         !
  REAL (KIND=DP)           :: wl_cut_esp        ! Specie ionisation cut off                !
  TYPE (RF_TYPE)           :: rf_esp            ! See above for the RF_TYPE definition     !
END TYPE PHOTO_ESP !------------------------------------------------------------------------

!===========================================================================================
! DECLARATIONS
!===========================================================================================
TYPE (PHOTO_ESP), PUBLIC, DIMENSION(:), ALLOCATABLE :: phdest               ! Table defined as a PHOTO_ESP TYPE variable

! Wavelength and Radiation field
TYPE (RF_TYPE), PUBLIC                             :: rf_wl                 !  Radiation field grid (WaveLength)
TYPE (RF_TYPE), PUBLIC                             :: rf_u                  !  Radiation field energy density (erg/cm^3/A)
TYPE (RF_TYPE), PUBLIC                             :: rf_incI               !  Incident radiation field specific intensity
                                                                            !  isotropic part, scaled by radm and radp
INTEGER, DIMENSION (:), POINTER, PUBLIC            :: wl_pointer            !  Reverse ordering of wavelength

REAL (KIND=dp), PUBLIC                             :: wlth0                 !  Start of Wavelength grid (H ionisation)
REAL (KIND=dp), PUBLIC                             :: wlthm                 !  End of Wavelength grid
REAL (KIND=dp), PUBLIC                             :: wl_hab = 2400.0_dp    !  Limit of Habing Field
REAL (KIND=dp), PUBLIC                             :: wl_6ev = 2066.37_dp   !  Limit of Habing Field
REAL (KIND=dp), PUBLIC                             :: pwlg = 16.0_dp        !  Initial WL step
INTEGER, PUBLIC                                    :: nwlg = 200            !  Initial number of WL points
INTEGER, PUBLIC                                    :: mfgk                  !  Number of lines (H2 + CO + ...) in FGK transfer
REAL (KIND=dp), PUBLIC, DIMENSION (:), ALLOCATABLE :: fgkwl                 !  Wavelength of "FGK lines"
REAL (KIND=dp), PUBLIC, DIMENSION (:), ALLOCATABLE :: fgkrf                 !  Line absorption (direct)
REAL (KIND=dp), PUBLIC, DIMENSION (:), ALLOCATABLE :: fgkrfr                !  Line absorption (reverse)
REAL (KIND=dp), PUBLIC, DIMENSION (:), ALLOCATABLE :: fgkgrd                !  Line self-shielding

! Photo-ionisation and photo-dissociation related quantities

REAL (KIND=dp), PUBLIC, DIMENSION (0:noptm)        :: pdish2                !  H2 Photdissociation rate at iopt
REAL (KIND=dp), PUBLIC, DIMENSION (nlevh2,0:noptm) :: pdh2vJ                !  H2 Photdissociation rate from level lev(v,J) at iopt
REAL (KIND=dp), PUBLIC, DIMENSION (nlevh2)         :: sdh2vJ                !  H2 Photdissociation rate from level lev(v,J) at iopt
REAL (KIND=dp), PUBLIC, DIMENSION (0:noptm)        :: pdishd                !  HD Photdissociation rate at iopt

REAL (KIND=dp), PUBLIC, DIMENSION (0:noptm)        :: pdico0                !  CO Photdissociation rate at iopt
REAL (KIND=dp), PUBLIC, DIMENSION (0:noptm)        :: pdico1                !  13CO Photdissociation rate at iopt
REAL (KIND=dp), PUBLIC, DIMENSION (0:noptm)        :: pdico2                !  C18O Photdissociation rate at iopt
REAL (KIND=dp), PUBLIC, DIMENSION (0:noptm)        :: pdico3                !  13C18O Photdissociation rate at iopt

REAL (KIND=DP), PUBLIC, DIMENSION(:,:), ALLOCATABLE:: pdiesp                !  Photoionisation rate at iopt of the studied speci

TYPE (RF_TYPE), PUBLIC                             :: rf_exter              !  External radiation field (if any)
INTEGER                                            :: nbpt, nskip

REAL (KIND=dp), PUBLIC                             :: int_rad0              !
REAL (KIND=dp), PUBLIC                             :: int_rad1              !
REAL (KIND=dp), PUBLIC                             :: int_rad2              !

! Grain photo-electric efficiency and extinction curve
TYPE(RF_TYPE), PUBLIC                              :: rf_extpou             !  Absorption by dust (Fitzpatrick)

REAL (KIND=dp), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: si_QU, si_QUl, si_QUl2
REAL (KIND=dp), PUBLIC, DIMENSION(:), ALLOCATABLE   :: si_
TYPE(RF_TYPE),PUBLIC, DIMENSION(:), ALLOCATABLE     :: sigabs
TYPE(RF_TYPE),PUBLIC                                :: dust_wl

! This is used for photo-electric heating (Bakes & Tielens formalism) in stv5c.f)
TYPE(RF_TYPE), public                               :: bt_int                ! Integrand (without sigma)

! This table saves the radiation field at some "imposed" Av for output
REAL (kind=dp), public, DIMENSION(:,:), ALLOCATABLE  :: rfield

INTEGER, PUBLIC                                      :: ic, ns
CHARACTER (LEN=15), DIMENSION(:), ALLOCATABLE        :: csect_files          ! List of reactions
CHARACTER (LEN=15), DIMENSION(:), ALLOCATABLE        :: les_esp              ! List of species
CHARACTER (LEN=15)                                   :: un_nom, de_nom
    
CONTAINS

!**********************************************************************************************************
  SUBROUTINE rf_init
!**********************************************************************************************************
! purpose : Initialise model dependant quantities related to Radiation Field (RF)
!   * Wavelength grid                  : rf_wl (need not have constant step)
!   * RF Energy Density                : rf_u (currently Draine field only)
!   * Dust absorption curve            : rf_extpou (Fitzpatrick fit)

!---------------------------Declaration of parameters-----------------------------------------------------
    USE constantes
    USE auxiliaries
    USE inoutput
    USE PROFIL
    USE DRAINE_DUST
    USE CHEM_DATA

    IMPLICIT NONE

    REAL (KIND=dp)                             :: wl_cut_D = 2000.0_dp !  Draine field cutoff wavelength
    INTEGER                                    :: I_cut_D              !  Draine field cutoff position in grid
    REAL (KIND=dp)                             :: wl_cut_F             !  Fitzpatrick cutoff wavelength
    INTEGER                                    :: I_cut_F              !  Fitzpatrick cutoff position in grid
    INTEGER                                    :: mm, k0
    INTEGER                                    :: num                  !  Dummy number
    INTEGER                                    :: i, j, k
    INTEGER                                    :: nb_pt                !  Number of points in a file
    INTEGER                                    :: np_before            !  Number of points before Lyman lim.
    INTEGER                                    :: add_1, iloc, iord    !  Used by rf_extand
    REAL (KIND=dp)                             ::  Energy, sigm
    REAL (KIND=dp)                             :: wl, wlo, rf, rfo
    TYPE (RF_TYPE)                             :: rf_tmp
    REAL (KIND=dp)                             :: summ_f, summ_f2
    INTEGER                                    :: npsec
    parameter (npsec = 100)

    REAL(KIND=dp)                              :: wl0, wl1, dopl, dwl
    REAL(KIND=dp), PARAMETER                   :: wlrinc = 1.2_dp
    INTEGER                                    :: njl, nvl
    INTEGER                                    :: jwlmin, jwlmax

    ! Declarations for intergalactic radiation field -------------------------------------------------------------
    ! Note : To use an intergalactic radiation field, the user has to provide the 3 following parameters
    !        Intergalactic spectra are read from haardt96.dat in the data directory
    !        To not use an intergalactic radiation field, program has to be compiled with F_haardt96 = 0
    INTEGER, PARAMETER          :: F_haardt96 = 0      ! FLAG 1 activate intergalactic radiation field
    INTEGER, PARAMETER          :: F_Haardt96_only = 0 ! FLAG 1 suppress Draine radiation field
                                                       !      0 adds intergal. spect. to Draine rad. field
    REAL (KIND=dp)              :: redshift = 2.43_dp   ! Redshift of the intergalactic rad. field
    ! Other quantities used to initialize the intergalactic spectrum
    INTEGER                     :: nredshift ! number of redshifts for which intergalactic spectra are provided
    INTEGER, PARAMETER          :: nlignmax_haardt96=1000 ! Max number of wavelength in haardt96.dat
    INTEGER                     :: nlign_haardt96  ! number of wavelength in haardt96.dat
    REAL (KIND=dp), ALLOCATABLE :: tab_redshift(:) ! table of known redshifts
    REAL (KIND=dp), ALLOCATABLE :: lam_haardt(:)   ! table of the wavelenght in Ang. in datafile 
    REAL (KIND=dp), ALLOCATABLE :: I_haardt(:,:)   ! spectra in function of wavelength and redshift
                                                   ! read in haardt96.dat in erg s-1 cm-2 A-1 str-1
    REAL (KIND=dp), ALLOCATABLE :: I_intergal(:)   ! spectrum at the proper redshift
    INTEGER                     :: num_redshift    ! number of the closest redshift in data to the one searched
    REAL (KIND=dp)              :: fact            ! factor used in linear interpolation

!-------------- Execution  ------------------------------------------------------------

    !==================================================================================
    ! INITIALIZATION of the radiation field grid
    !==================================================================================
    ! Compute H ionisation potential --------------------------------------------------
    hiop  = 2.0_dp * xpi*xpi * xqe**4 * xme / (xhp*xhp)
    wlth0 = 1.0e8_dp * xhp * clum / hiop
    print *, " true wlth0 =", wlth0
    wlth0 = praih(nraih,2)
    wlthm = wlth0 + nwlg * pwlg
    print *, "  wlth0 =", wlth0, " wlthm =", wlthm

    ! Initialisation of wavelengths ---------------------------------------------------
    ! rf_wl%Ord is a permutation of (rf_wl%lower : rf_wl%upper) giving the
    !           order in which rf_wl%Val have been entered

    CALL rf_create (rf_wl, 0, nwlg)
    rf_wl%Val = wlth0 + pwlg * (/ (I, I=0,nwlg) /)
    rf_wl%Ord = (/ (I, I=0,nwlg) /)

    !================================================================================
    ! Add all "special values"
    !================================================================================

    !---  Habing field limit --------------------------------------------------------
    CALL rf_extand (rf_wl, wl_hab, add_1, iloc, iord)
    nwlg = nwlg + add_1

    !--- 6 eV field limit -----------------------------------------------------------
    CALL rf_extand (rf_wl, wl_6ev, add_1, iloc, iord)
    nwlg = nwlg + add_1

    !--- Fitzpatrick & Massa extinction curve ---------------------------------------
    wl_cut_F = 1.0e4_dp / 5.9_dp
    CALL rf_extand (rf_wl, wl_cut_F, add_1, iloc, iord)
    nwlg = nwlg + add_1
    !   print *, iloc, iord, wl_cut_F

    !--- Draine Field cut-off ------------------------------------------------------
    CALL rf_extand (rf_wl, wl_cut_D, add_1, iloc, iord)
    nwlg = nwlg + add_1
    !   print *, iloc, iord, wl_cut_D
    !   print *, 'nwlg after :', nwlg

    !--- Photodestruction sections in data/section --------------------------------

    ! Loop on all photodestruction reactions computed with cross sections
    do k = 1, nphot_i - i_mandat

       fichier = TRIM(data_dir)//'sections/'//TRIM(csect_files(k))//".dat"
       WRITE(iscrn,*) "Ouverture de : ",fichier
       OPEN(iread2, file = fichier, status = 'old')

       ! First count number of point above threshold
       READ(iread2,"(2/)")
       READ(iread2,*) nb_pt  ! number of points in the file
       READ(iread2,*)

       ! Remove points below Lyman limit
       np_before = 0
       DO i = 1, nb_pt
          READ (iread2,*) num, wl ! Energy or wavelength in Angstrom
          if (wl < wlth0) then
             np_before = np_before + 1
          endif
       END DO
       nb_pt = nb_pt - np_before

       ! close(iread2)

       ! Then read correct number of cross section and interpolate last to wlth0
       CALL rf_create (phdest(k)%rf_esp,1,nb_pt+1)

       ! OPEN(iread2, file = fichier, status = 'old')
       REWIND(iread2)
       READ(iread2,"(4/)")

       if (np_before > 0) then ! jump points to remove
          do i=1,np_before
             READ(iread2,*) ! Excluded photo destruction data
          end do
       end if

       DO i = 1, nb_pt ! store data
          READ (iread2,*) num, Energy, sigm
          wl = Energy
          if (i == nb_pt) then
             phdest(k)%wl_cut_esp = wl
          endif

          CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
          nwlg = nwlg + add_1
          phdest(k)%rf_esp%Val(i+1) = sigm   ! sigm's unit is cm^-2
          phdest(k)%rf_esp%Ord(i+1) = iord
       END DO
       CLOSE(iread2)

       ! Eventual interpolation at wlth0
       ! Default : If no value is given for section at wlth0, the section
       ! at the next point is used
       if (phdest(k)%r1 == 's') then
          phdest(k)%rf_esp%Val(1) = phdest(k)%rf_esp%Val(2) !Possible interpolation here 
       else
          phdest(k)%rf_esp%Val(1) = phdest(k)%rf_esp%Val(2)
       end if
       CALL rf_locate (rf_wl, wlth0, i)
       phdest(k)%rf_esp%Ord(1) = rf_wl%Ord(i)

       CALL rf_revers (wl_pointer)
       if (rf_wl%Val(nwlg) > wlthm) then
         wlthm = rf_wl%Val(nwlg)
       endif
    end do

    !--- External radiation field -----------------------------------------

    if (srcpp(1:2) == 'F_') then
       print *, srcpp
       fichier = TRIM(data_dir)//srcpp
       WRITE(iscrn,*) "Ouverture de : ", fichier
       open(iread2, file = fichier, status = 'old')
       read (iread2,*) rstar
       read (iread2,*) teff
       print *, rstar, teff
       rstar = rstar * rsun
       read (iread2,*) nb_pt
       read (iread2,*)
       print *, nb_pt

       nbpt = 0
       nskip = 0
       do i =1,nb_pt
          read (iread2,*) wl
          wl = wl * 1.0e1_dp      ! convert to Angstrom
          if (wl < wlth0) then
             nskip = nskip + 1
             cycle
          endif
          nbpt = nbpt + 1
          if (wl > wlthm) then
             exit
          endif
       end do
       close (iread2)
       nbpt = nbpt + 1
       print *, nbpt

       call rf_create (rf_exter,1,nbpt)
       open(iread2, file = fichier, status = 'old')
       read(iread2,"(3/)")
       wl = 0.0_dp
       rf = 0.0_dp
       do i=1,nskip+1
          wlo = wl
          rfo = rf
          !  Final value of rf must be in erg cm-2 s-1 A-1 sterad-1 (i.e. a specific intensity)
          read (iread2,*) wl, rf
          wl = wl * 10._dp      ! convert to Angstrom
          rf = rf * 0.1_dp      ! convert to Angstr-1
       end do

       CALL rf_locate (rf_wl, wlth0, i)
       rf_exter%Val(1) = rfo + (rf - rfo) * &
                         (wlth0 - wlo) / (wl - wlo)
       rf_exter%Ord(1) = rf_wl%Ord(i)
       CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
       nwlg = nwlg + add_1
       rf_exter%Val(2) = rf
       rf_exter%Ord(2) = iord

       do i =3,nbpt
          wlo = wl
          rfo = rf
          read (iread2,*) wl, rf
          wl = wl * 10.0_dp      ! convert to Angstrom
          rf = rf * 0.1_dp       ! convert to Angstr-1
          if (wl > wlthm) then
             exit
          endif
          CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
          nwlg = nwlg + add_1
          rf_exter%Val(i) = rf
          rf_exter%Ord(i) = iord
       end do
       close (iread2)

       CALL rf_locate (rf_wl, wlthm, i)
       rf_exter%Val(nbpt) = rfo + (rf - rfo) * &
                            (wlthm - wlo) / (wl - wlo)
       rf_exter%Ord(nbpt) = rf_wl%Ord(i)

    endif

    !--- Grains absorption cross sections ----------------
    !    (interpolated from Draines data for each radius)

    CALL READ_DRAINE_FILES

    CALL MIXGRSI (wlth0, wlthm, jwlmin, jwlmax)

    ALLOCATE(sigabs(npg))
    do mm=1,npg
       NULLIFY (sigabs(mm)%val,sigabs(mm)%ord)                     !SG
       CALL rf_create(sigabs(mm),1,jwlmax-jwlmin+2)
    end do

    DO i=1,jwlmax-jwlmin+2
       if (i == 1) then
          wl = wlth0
       else if (i == jwlmax-jwlmin+2) then
          wl = wlthm
       else
          wl = wavgrd(i+jwlmin-1)
       endif
       CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
       nwlg = nwlg + add_1
       do mm=1,npg
          sigabs(mm)%Val(i) = Qabs(i,mm)
          sigabs(mm)%Ord(i) = iord
       end do
    END DO

    CALL rf_revers (wl_pointer)

    !----------------------------------------------------------------------
    !  Extension of the wavelength grid
    !----------------------------------------------------------------------
    ! The wavelength grid need now be extended to give accurate integration
    ! 3 -  integrate test function over wavelength

    ! Search for sulfur photo-ionization
    do k = 1, nphot_i - i_mandat
       if (phdest(k)%r1 == "s") then
          k0 = k
          exit
       endif
    end do
    ! Compute sulfur ionization rate
    summ_f = sect_int (phdest(k0)%rf_esp, rf_wl, wl_pointer, phdest(k0)%wl_cut_esp)
    !print *, "  summ_f :", summ_f

    ! 4 -  iteratively refine wavelength grid until continuum integration is "good"
    summ_f2 = 0.5_dp * summ_f
    j = 0
    ! print *, "   nwlg 1:", nwlg
    do while (abs((summ_f-summ_f2)/summ_f) > 1.0e-5)
       j = j + 1

       pwlg = pwlg * 0.5_dp
       i = 0
       do while (i < nwlg-1)
          if ((rf_wl%Val(i+1)-rf_wl%Val(i)) > pwlg) then
             wl = 0.5_dp * (rf_wl%Val(i+1)+rf_wl%Val(i))
             CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
             nwlg = nwlg + add_1
          else
             i = i + 1
          endif
       end do
       print *, "   nwlg",j,":", nwlg

       CALL rf_revers (wl_pointer)
       summ_f2 = summ_f
       summ_f = sect_int (phdest(k0)%rf_esp, rf_wl, wl_pointer, phdest(k0)%wl_cut_esp)
       ! print *, "  summ_f :", summ_f2, summ_f

    end do

    !---------------------------------------------------------------------------
    ! ADD LINE WAVELENGTH IN THE GRID
    !---------------------------------------------------------------------------

    !--- Atomic hydrogen (Lyman lines) -----------------------------------------
    !     parameter "dopl" is the step in lambda mesured in Doppler width
    if (jlyman == 1) then
       do i=1,nraih
          wl0 = praih(i,2)
          dopl = 0.3_dp * wl0 * vturb / clum
          CALL rf_extand (rf_wl, wl0, add_1, iloc, iord)
          nwlg = nwlg + add_1

          dwl = 0.2_dp * dopl
          wl = wl0
          do
             if (abs(wl-wl0)/dopl < wlrwid) then
                wl1 = rf_wl%Val(iloc+1)
                if ((wl1-wl) > 2.0_dp*dwl) then
                   wl = wl + dwl
                   CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                else if ((wl1-wl) > dwl) then
                   wl = wl + (wl1-wl) * 0.5_dp
                   CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc + 1
                else
                   wl = wl1
                   iloc = iloc + 1
                endif
                dwl = wlrinc * dwl
             else
                exit
             endif
          end do

          call rf_locate (rf_wl, wl0, iloc)

          dwl = 0.2_dp * dopl
          wl = wl0
          do
             if (abs(wl-wl0)/dopl < wlrwid .AND. &
                wl > wlth0) then
                wl1 = rf_wl%Val(iloc-1)
                if ((wl-wl1) > 2.0_dp*dwl) then
                    wl = wl - dwl
                    CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                    nwlg = nwlg + add_1
                else if ((wl-wl1) > dwl) then
                    wl = wl - (wl-wl1) * 0.5_dp
                    CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                    nwlg = nwlg + add_1
                    wl = wl1
                    iloc = iloc - 1
                else
                    wl = wl1
                    iloc = iloc - 1
                endif
                dwl = wlrinc * dwl
             else
                exit
             endif
          end do

       end do
    endif

    !---  CO lines , only lowest J --------------------------------

    do i=1,nrco0
       njl = irco0(i)
       if (njl >= jfgkco0) then
          cycle
       endif
       wl0 = prco0(i,2)
       dopl = 0.3_dp * wl0 * vturb / clum
       CALL rf_extand (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1

       dwl = 0.2_dp * dopl
       wl = wl0
       do
          if (abs(wl-wl0)/dopl < wlrwid) then
             wl1 = rf_wl%Val(iloc+1)
             if ((wl1-wl) > 2.0_dp*dwl) then
                wl = wl + dwl
                CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
             else if ((wl1-wl) > dwl) then
                wl = wl + (wl1-wl) * 0.5_dp
                CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
                wl = wl1
                iloc = iloc + 1
             else
                wl = wl1
                iloc = iloc + 1
             endif
             dwl = wlrinc * dwl
          else
             exit
          endif
       end do

       call rf_locate (rf_wl, wl0, iloc)

       dwl = 0.2_dp * dopl
       wl = wl0
       do
          if (abs(wl-wl0)/dopl < wlrwid .AND. &
             wl > wlth0) then
             wl1 = rf_wl%Val(iloc-1)
             if ((wl-wl1) > 2.0_dp*dwl) then
                wl = wl - dwl
                CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
             else if ((wl-wl1) > dwl) then
                wl = wl - (wl-wl1) * 0.5_dp
                CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
                wl = wl1
                iloc = iloc - 1
             else
                wl = wl1
                iloc = iloc - 1
             endif
             dwl = wlrinc * dwl
          else
             exit
          endif
       end do
 
    end do

    !--- H2 lines , only lowest J -------------------------------------

    do i = 1, nrh2bt
       nvl = irah2b(i,1)
       njl = irah2b(i,2)
       if (nvl > 0 .OR. njl >= jfgkh2) then
          cycle
       endif
       wl0 = prah2b(i,2)
       dopl = 0.3_dp * wl0 * vturb / clum
       CALL rf_extand (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1

       dwl = 0.2_dp * dopl
       wl = wl0
       do
          if (abs(wl-wl0)/dopl < wlrwid) then
             wl1 = rf_wl%Val(iloc+1)
             if ((wl1-wl) > 2.0_dp*dwl) then
                wl = wl + dwl
                CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
             else if ((wl1-wl) > dwl) then
                wl = wl + (wl1-wl) * 0.5_dp
                CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
                wl = wl1
                iloc = iloc + 1
             else
                wl = wl1
                iloc = iloc + 1
             endif
             dwl = wlrinc * dwl
          else
             exit
          endif
       end do

       call rf_locate (rf_wl, wl0, iloc)

       dwl = 0.2_dp * dopl
       wl = wl0
       do 
          if (abs(wl-wl0)/dopl < wlrwid .AND. &
             wl > wlth0) then
             wl1 = rf_wl%Val(iloc-1)
             if ((wl-wl1) > 2.0_dp*dwl) then
                wl = wl - dwl
                CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
             else if ((wl-wl1) > dwl) then
                wl = wl - (wl-wl1) * 0.5_dp
                CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
                wl = wl1
                iloc = iloc - 1
             else
                wl = wl1
                iloc = iloc - 1
             endif
             dwl = wlrinc * dwl
          else
             exit
          endif
       end do

    end do

    do i=1,nrh2ct
      nvl = irah2c(i,1)
      njl = irah2c(i,2)
      if (nvl > 0 .OR. njl >= jfgkh2) then
        cycle
      endif
      wl0 = prah2c(i,2)
      dopl = 0.3_dp * wl0 * vturb / clum
      CALL rf_extand (rf_wl, wl0, add_1, iloc, iord)
      nwlg = nwlg + add_1

      dwl = 0.2_dp * dopl
      wl = wl0
      do
        if (abs(wl-wl0)/dopl < wlrwid) then
          wl1 = rf_wl%Val(iloc+1)
          if ((wl1-wl) > 2.0_dp*dwl) then
            wl = wl + dwl
            CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
            nwlg = nwlg + add_1
          else if ((wl1-wl) > dwl) then
            wl = wl + (wl1-wl) * 0.5_dp
            CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
            nwlg = nwlg + add_1
            wl = wl1
            iloc = iloc + 1
          else
            wl = wl1
            iloc = iloc + 1
          endif
          dwl = wlrinc * dwl
        else
          exit
        endif
      end do

      call rf_locate (rf_wl, wl0, iloc)

      dwl = 0.2_dp * dopl
      wl = wl0
      do
        if (abs(wl-wl0)/dopl < wlrwid .AND. &
            wl > wlth0) then
          wl1 = rf_wl%Val(iloc-1)
          if ((wl-wl1) > 2.0_dp*dwl) then
            wl = wl - dwl
            CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
            nwlg = nwlg + add_1
          else if ((wl-wl1) > dwl) then
            wl = wl - (wl-wl1) * 0.5_dp
            CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
            nwlg = nwlg + add_1
            wl = wl1
            iloc = iloc - 1
          else
            wl = wl1
            iloc = iloc - 1
          endif
          dwl = wlrinc * dwl
        else
          exit
        endif
      end do

    end do

    print *, "   nwlg =", nwlg
!   print *, "   pwlg =", pwlg

    if (rf_wl%upper /= nwlg) then
      print *, "Probleme"
    endif

    CALL rf_revers (wl_pointer)

    !-------------------------------------------------------------------------------------
    !****    From here, the set of wavelength is complete and should not be modified ****
    !-------------------------------------------------------------------------------------

    !-------------------------------------------------------------------------------------
    ! Initialisation of the radiation Field using Draine Formula as revised by M. Kopp
    !-------------------------------------------------------------------------------------

    CALL rf_create (rf_u, 0, nwlg)
    CALL rf_create (rf_incI, 0, nwlg)

    ! Draine Formula: Specific intensity
    ! Radiation field in erg cm-2 s-1 A-1 sterad-1

    CALL rf_locate (rf_wl, wl_cut_D, I_cut_D)

    ! Draine 1978 radiation field
    ! Expression from M. Kopp PhD thesis page 284, from Sternberg and Dalgarno 1985.
    rf_incI%Val(0:I_cut_D) = 6.36E7_dp / rf_wl%Val(0:I_cut_D)**4 &
                           - 1.0237E11_dp / rf_wl%Val(0:I_cut_D)**5 &
                           + 4.0812E13_dp / rf_wl%Val(0:I_cut_D)**6

    rf_incI%Val(I_cut_D+1:nwlg) = 1.38243E-5_dp &
                                * rf_wl%Val(I_cut_D+1:nwlg)**(-0.3_dp)

    rf_incI%Ord = (/ (I, I=rf_incI%lower,rf_incI%upper) /)
    rf_incI%Val = rf_incI%Val / (4.0 * xpi) ! rf_incI is now in erg cm-2 s-1 str-1 A-1 

    !----------------------------------------------------------------------------------------
    ! Initialization of an intergalactic radiation field (Haardt & Madau 1996, ApJ, 461, 20)
    !----------------------------------------------------------------------------------------
    IF (f_Haardt96 == 1) THEN
 
       IF (f_Haardt96_only == 1) THEN
          ! On ne prend en compte que Haardt & Madau 96
          ! aussi on commence par "effacer" Draine
          rf_incI%Val(0:nwlg) = 0.0_dp
          ! Sinon, le champ de intergalactique viendra s'ajouter a Draine
       ENDIF
      
       ! Modification of the temperature of the CMB
       tcmb = 2.73_dp * (1.0_dp + redshift)

       fichier = TRIM(data_dir)//TRIM("haardt96.dat")
       WRITE(iscrn,"('Ouverture de :',A30)") fichier
       OPEN (10, file = fichier, status="old", action="read")
       READ(10,*)           ! saute ligne de commentaire
       READ(10,*)           ! saute ligne de commentaire
       READ(10,*) nredshift ! number of redshifts in table
       ! Allocation de memoire pour stocker les donnees de la table
       ALLOCATE(tab_redshift(nredshift))
       ALLOCATE (lam_haardt(nlignmax_haardt96))
       ALLOCATE (I_haardt(nlignmax_haardt96,nredshift))  
       ALLOCATE(I_intergal(nlignmax_haardt96))     
       ! Cherche la liste des redshifts disponibles dans la table
       READ(10,*) (tab_redshift(i), i=1, nredshift)
       
       !Lecture des donnees
       DO i = 1, nlignmax_haardt96
          READ(10,*,END=110) lam_haardt(i), (I_haardt(i,k), k=1, nredshift)
!          PRINT *,i,lam_haardt(i),I_haardt(i,26)
       ENDDO
110    continue
       nlign_haardt96 = i - 1 ! number of wavelength in haardt96.dat
       CLOSE(10)

       ! Search in haardt96.dat for an interval of redshift containing the whished redshift
       ! If searched redshift outside of data :
       IF (redshift < tab_redshift(1)) THEN
          PRINT *,"Redshift too small compared to data"
          PRINT *,"Switch to smallest redshift in data :", tab_redshift(1)
          redshift = tab_redshift(1)
       ENDIF
       IF (redshift > tab_redshift(nredshift)) THEN
          PRINT *,"Redshift too big compared to data"
          PRINT *,"Switch to highest redshift in data :",tab_redshift(nredshift)
          redshift = tab_redshift(nredshift)
       ENDIF
       ! Searched redshift inside data range
       ! On traite les cas particuliers :
       ! Si le redshift cherche est le premier de la liste
       IF (redshift == tab_redshift(1)) THEN
          I_intergal(:) = I_haardt(:,1)
       ENDIF
       ! Si le redshift cherche est le dernier de la liste
       IF (redshift == tab_redshift(nredshift)) THEN
          I_intergal(:) = I_haardt(:,nredshift)
       ENDIF
       ! Cas general
       IF ( tab_redshift(1) < redshift .AND. redshift < tab_redshift(nredshift)) THEN
          num_redshift = 1
          DO WHILE (redshift > tab_redshift(num_redshift))
             num_redshift = num_redshift + 1
             PRINT *,num_redshift, redshift, tab_redshift(num_redshift), tab_redshift(num_redshift+1)
          ENDDO
          ! Interpolation lineaire du redshift
          IF (redshift == tab_redshift(num_redshift-1)) THEN
             I_intergal(:) = I_haardt(:,num_redshift-1)
             PRINT *,"Exact redshift found in data :",num_redshift,redshift, tab_redshift(num_redshift-1)
          ELSE IF (redshift == tab_redshift(num_redshift)) THEN
             I_intergal(:) = I_haardt(:,num_redshift)
             PRINT *,"Exact redshift found in data :",num_redshift,redshift, tab_redshift(num_redshift)
          ELSE IF ((tab_redshift(num_redshift-1) < redshift) .AND. (redshift < tab_redshift(num_redshift))) THEN
             fact = (redshift - tab_redshift(num_redshift-1)) / (tab_redshift(num_redshift) - tab_redshift(num_redshift-1))
             I_intergal(:) = I_haardt(:,num_redshift-1) * (1.0_dp-fact) + I_haardt(:,num_redshift) * fact
             PRINT *,"Interpolation of intergal. spect. between z :",tab_redshift(num_redshift-1), tab_redshift(num_redshift)
          ENDIF
       ENDIF
       
       ! Remplissage du tableau des intensites avec interpolation lineaire aux lam. necessaires
       k = 0
       DO WHILE (rf_wl%Val(k) < lam_haardt(1))
          PRINT *,"Warning : Data for intergal. spectrum do not begin at a small enough lambda"
          PRINT *,"          Extrapolation"
          rf_incI%Val(k) = rf_incI%Val(k) + I_intergal(1)
          k = k + 1
       ENDDO
       j = 1
       DO i = k, nwlg
          DO WHILE (rf_wl%Val(i) > lam_haardt(j)) 
             j = j + 1
          ENDDO
          IF (lam_haardt(j) == rf_wl%Val(i)) THEN
             rf_incI%Val(i) = rf_incI%Val(i) + I_intergal(j)
          ELSE IF ((lam_haardt(j-1) < rf_wl%Val(i)) .AND. (rf_wl%Val(i) < lam_haardt(j))) THEN
             ! On interpole en lambda dans la table haardt.dat 
             ! interpolation pour le premier redshift
             fact = (rf_wl%Val(i) - lam_haardt(j-1)) / (lam_haardt(j) - lam_haardt(j-1))
             rf_incI%Val(i) = rf_incI%Val(i) + I_intergal(j-1) * (1.0_dp - fact) &
                                             + fact * I_intergal(j)
          ELSE IF (rf_wl%Val(i) > lam_haardt(nlign_haardt96)) THEN
             rf_incI%Val(i) = rf_incI%Val(i) + I_intergal(nlign_haardt96)
             PRINT *,"WARNING : Not enough point in lambda in haardt96.dat"
             PRINT *,"          Lambda searched :", rf_wl%Val(i)
          ENDIF  
       ENDDO
!      VERIFICATION DU CHAMP 
!       OPEN (10,file="toto.dat",status="unknown",action="write")
!       DO i = 0,nwlg
!          WRITE(10,*) rf_wl%Val(i), rf_incI%Val(i)
!       ENDDO
!       CLOSE(10)
!       PRINT *,"Fin d'ecriture du champ initial dans toto.dat"
!       PAUSE
             
       ! Remarque : Pas besoin de toucher a rf_incI%Ord. Il a ete correctement construit lors
       !            du remplissage du champ de Draine. 

       DEALLOCATE(tab_redshift)
       DEALLOCATE(lam_haardt)
       DEALLOCATE(I_haardt)
       DEALLOCATE(I_intergal)

    ENDIF


    !----------------------------------------------------------------------------------
    ! Fitzpatrick extinction curves
    !----------------------------------------------------------------------------------

    CALL rf_create (rf_extpou, 0, nwlg)
    CALL rf_create (rf_tmp, 0, nwlg)
    rf_tmp%Val = 1.0e4_dp / rf_wl%Val

    CALL rf_locate (rf_wl, wl_cut_F, I_cut_F)

    rf_extpou%Val =  cc1 + cc2 * rf_tmp%Val + cc3 * (rf_tmp%Val * rf_tmp%Val / &
       ((rf_tmp%Val * rf_tmp%Val - y0 * y0)**2 + xgam * xgam * rf_tmp%Val * rf_tmp%Val))

    rf_extpou%Val(0:I_cut_F) = rf_extpou%Val(0:I_cut_F) &
       +  cc4 * (0.539_dp * (rf_tmp%Val(0:I_cut_F)-5.9_dp)**2 &
       + 0.0564_dp * (rf_tmp%Val(0:I_cut_F)-5.9_dp)**3)

    rf_extpou%Val = 1.0_dp + rf_extpou%Val / rrr
    rf_extpou%Ord = (/ (I, I=rf_extpou%lower,rf_extpou%upper) /)

    CALL rf_free (rf_tmp)

    !---------------------------------------------------------------------
    ! Create RF_TYPE objects and tables used in fbth (stv5c.f)
    !---------------------------------------------------------------------

    ALLOCATE (si_(rf_wl%lower:rf_wl%upper))
    ALLOCATE (si_QU(npg,rf_wl%lower:rf_wl%upper))
    ALLOCATE (si_QUl(npg,rf_wl%lower:rf_wl%upper))
    ALLOCATE (si_QUl2(npg,rf_wl%lower:rf_wl%upper))

    CALL rf_create (rf_tmp, 0, nwlg)
    rf_tmp%Val = rf_wl%Val * rf_incI%Val * J_to_u

    do mm=1,npg
       rf_tmp%Ord = rf_incI%Ord
       rf_tmp%Val = rf_incI%Val * J_to_u
       call sect_pri(sigabs(mm), rf_tmp, wl_pointer, si_, nwlg)
       si_QU(mm,:) = si_
       rf_tmp%Val = rf_tmp%Val * rf_wl%Val
       call sect_pri(sigabs(mm), rf_tmp, wl_pointer, si_, nwlg)
       si_QUl(mm,:) = si_
       rf_tmp%Val = rf_tmp%Val * rf_wl%Val
       call sect_pri(sigabs(mm), rf_tmp, wl_pointer, si_, nwlg)
       si_QUl2(mm,:) = si_
    end do

    !-----------------------------------------------------------
    ! Create Radiation field table for output only
    !-----------------------------------------------------------

    ALLOCATE (rfield(0:n_imp,rf_wl%lower:rf_wl%upper))

    CALL rf_free (rf_tmp)



END SUBROUTINE rf_init

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE sect_pri (sect, rf_fn, rindex, sect_p, nwlg)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Integrate The product sect * rf_fn over lambda
!           and save each partial sum from lamb1 to lamb_i

  IMPLICIT NONE

  INTEGER                             :: nwlg
  TYPE(RF_TYPE), intent(in)           :: sect       ! Section to integrate
  TYPE(RF_TYPE), intent(in)           :: rf_fn      ! Weight function
  INTEGER, POINTER, DIMENSION(:)      :: rindex     ! Reverse ordering index
  REAL (KIND = dp), DIMENSION(0:nwlg) :: sect_p     ! partial integrated section

  INTEGER                        :: i, j
  INTEGER                        :: ll, uu
  REAL (KIND = dp)               :: s1, s2, l1, l2
  REAL (KIND = dp)               :: s, l, so, lo

  sect_p = 0.0_dp
  sect_p(rindex(sect%Ord(sect%lower))) = 1.0e-40_dp

  do i=sect%lower, sect%upper-1
    ll = rindex(sect%Ord(i))
    l1 = rf_wl%Val(ll)
    s1 = sect%Val(i)

    uu = rindex(sect%Ord(i+1))
    l2 = rf_wl%Val(uu)
    s2 = sect%Val(i+1)

    lo = l1
    so = s1
    do j=ll,uu-1
      l = rf_wl%Val(j+1)
      s = s1 + (s2 - s1) * (l - l1) / (l2 - l1)
      sect_p(j+1) = sect_p(j) + 0.5_dp * (l - lo) &
                  * (so * rf_fn%Val(j) + s * rf_fn%Val(j+1))
      lo = l
      so = s
    end do

  end do

  END SUBROUTINE sect_pri

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  REAL(KIND=dp) FUNCTION  sect_int (sect, rf_fn, rindex, wl_thresh)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!---------------------------------------------------------------------------------------------
! Integrate The product sect * rf_fn over lambda
!           From lamb1 to lamb2
! sect is interpolated between two successive points
! rf_fn must be known at every point in lambda
!---------------------------------------------------------------------------------------------

  IMPLICIT NONE

  TYPE(RF_TYPE), intent(in)      :: sect       ! Section to integrate
  TYPE(RF_TYPE), intent(in)      :: rf_fn      ! Weight function
  INTEGER, POINTER, DIMENSION(:) :: rindex     ! Reverse ordering index
  REAL (KIND = dp)               :: wl_thresh  ! Threshold

  INTEGER                        :: i, ll, uu
  INTEGER                        :: I_thresh
  REAL (KIND = dp)               :: sumt, sump
  REAL (KIND = dp)               :: s1, s2, l1, l2
  REAL (KIND = dp)               :: i_int, j_int
  TYPE (RF_TYPE)                 :: rf_xx
  TYPE (RF_TYPE)                 :: rf_fxx

  call rf_locate (rf_wl, wl_thresh, I_thresh)

  sumt = 0.0_dp
  do i=sect%lower, sect%upper-1

    ll = rindex(sect%Ord(i))
    l1 = rf_wl%Val(ll)
    s1 = sect%Val(i)

    uu = rindex(sect%Ord(i+1))
    l2 = rf_wl%Val(uu)
    s2 = sect%Val(i+1)

    if (uu > I_thresh) then
      uu = I_thresh
      s2 = s1 + (s2 - s1) * (rf_wl%Val(uu) - l1) / (l2 - l1)
      l2 = rf_wl%Val(uu)
    endif

    if (l2 > l1) then

      CALL rf_create (rf_xx, ll, uu)
      CALL rf_create (rf_fxx, ll, uu)

      rf_xx%Val = rf_wl%Val(ll:uu)
      rf_fxx%Val = rf_fn%Val(ll:uu)
      i_int = rf_integration (rf_xx, rf_fxx)
      rf_fxx%Val = rf_fxx%Val * rf_xx%Val
      j_int = rf_integration (rf_xx, rf_fxx)

      sump = (j_int * (s1-s2) - i_int * (s1*l2-s2*l1)) / (l1-l2)    ! Analytical integration of sect
      sumt = sumt + sump
      CALL rf_free (rf_fxx)
      CALL rf_free (rf_xx)
    else
      exit
    endif

  end do

  sect_int = sumt

  END FUNCTION sect_int

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  REAL(KIND=dp) FUNCTION  rf_integration (xx, ffx)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!---------------------------------------------------------------------------------------------
! Integrate Function using the trapezium rules.
! With uneven step
!---------------------------------------------------------------------------------------------

  IMPLICIT NONE

  TYPE(RF_TYPE), intent(in) :: xx               ! Abscissa
  TYPE(RF_TYPE), intent(in) :: ffx              ! Y-axis
  INTEGER                   :: l, u, i
  l = xx%lower
  u = xx%upper

  rf_integration = 0.5_dp * SUM( (/ ((xx%Val(i+1) - xx%Val(i)) &
                 * (ffx%Val(i+1) + ffx%Val(i)), i= l, u-1) /) )

  END FUNCTION rf_integration

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE rf_create (rf_xx, ll, uu)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Purpose: initialize upper and lower range of Val and Ord
!          and allocate the corresponding vector
! Note : Do NOT forget to deallocate temporary rf_objects !

  USE constantes

  IMPLICIT NONE

  TYPE(RF_TYPE), intent(inout) :: rf_xx
  INTEGER, intent (in)         :: ll, uu

  rf_xx%lower = ll
  rf_xx%upper = uu
!Choisir la ligne en fonction du compilateur
!pour pgi: allocated et pour absoft: associated
!Sous Windows (Visual Fortran) : associated
! IF ( .NOT. ALLOCATED(rf_xx%Val)) THEN
  IF ( .NOT. ASSOCIATED(rf_xx%Val)) THEN
    ALLOCATE (rf_xx%Val(ll:uu))
    ALLOCATE (rf_xx%Ord(ll:uu))
  ELSE
!   PRINT *, "  Warning : Allocation is done by rf_create"
!   PRINT *, "            Do not forget to deallocate if called from a loop"
!   CALL rf_free (rf_xx)
    ALLOCATE (rf_xx%Val(ll:uu))
    ALLOCATE (rf_xx%Ord(ll:uu))
  ENDIF

  rf_xx%Val = 0.0_dp
  rf_xx%Ord = 0

  END SUBROUTINE rf_create

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE rf_free (rf_xx)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Purpose: deallocate all allocatable bits of rf_xx

  IMPLICIT NONE
  TYPE(RF_TYPE), INTENT (INOUT) :: rf_xx

  DEALLOCATE (rf_xx%Val)
  DEALLOCATE (rf_xx%Ord)

  END SUBROUTINE rf_free

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE rf_revers (rindex)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Purpose: Reverse ordering of wavelength

  IMPLICIT NONE
  INTEGER, POINTER, DIMENSION(:) :: rindex
  INTEGER :: i

  NULLIFY (rindex)
  ALLOCATE (rindex(rf_wl%lower:rf_wl%upper))

  do i=rf_wl%lower,rf_wl%upper
    rindex(rf_wl%Ord(i)) = i
  end do

  END SUBROUTINE rf_revers

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE rf_extand (rf_xx, x, add_1, iloc, iord)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Purpose: Add one value in rf_xx%Val if not already here
!          and update upper and rf_xx%Ord
! On output, iloc is the index of the added value in rf_xx
!            iord is the rank (order) of the added value

  IMPLICIT NONE

  TYPE(RF_TYPE), intent(inout) :: rf_xx
  REAL (KIND=dp), intent (in)  :: x
  INTEGER, INTENT (out)        :: add_1
  INTEGER, INTENT (out)        :: iloc
  INTEGER, INTENT (out)        :: iord

  TYPE(RF_TYPE)                :: rf_tmp
  INTEGER                      :: ll, uu
  INTEGER                      :: i

  ll = rf_xx%lower
  uu = rf_xx%upper

  CALL rf_locate (rf_xx, x, i)

! if x is already in rf_xx%Val, no need to extand

  IF (rf_xx%Val(i) == x) THEN
    add_1 = 0
    iloc = i
    iord = rf_xx%Ord(i)
    RETURN
  ENDIF

  CALL rf_create (rf_tmp, ll, uu)

  rf_tmp%lower = ll
  rf_tmp%upper = uu
  rf_tmp%Val = rf_xx%Val
  rf_tmp%Ord = rf_xx%Ord

  CALL rf_free (rf_xx)
  CALL rf_create (rf_xx, ll, uu+1)

  rf_xx%Val(ll:i) = rf_tmp%Val(ll:i)
  rf_xx%Ord(ll:i) = rf_tmp%Ord(ll:i)

  iloc = i + 1
  iord = uu + 1
  rf_xx%Val(iloc) = x
  rf_xx%Ord(iloc) = iord

  rf_xx%Val(i+2:uu+1) = rf_tmp%Val(i+1:uu)
  rf_xx%Ord(i+2:uu+1) = rf_tmp%Ord(i+1:uu)

  rf_xx%upper = uu + 1

  add_1 = 1
  CALL rf_free (rf_tmp)

  END SUBROUTINE rf_extand

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE Shell_sort (rf_xxin,rf_xxout)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!----------------------------------------------------------------------------------------------
! purpose: sort an rf_object in increasing ordre.
!      The sort methode is the "Shell" sort
!      Num Rec in Fortran, p 322  (second edition)
! On input rf_xxout%lower and rf_xxout%upper should have been
!      set to check compatibility of objects
! Outside Shell_sort, rf_xxin and rf_xxout can be the same object
!      In that case the sorted one replaces the unsorted one.
!----------------------------------------------------------------------------------------------

  IMPLICIT NONE

  TYPE(RF_TYPE), INTENT(IN)       :: rf_xxin
  TYPE(RF_TYPE), INTENT(INOUT)    :: rf_xxout

  TYPE(RF_TYPE)                   :: a
  INTEGER                         :: p
  REAL(KIND=dp)                   :: v
  INTEGER                         :: i, j, inc, n, l, u

!------------------------Execution------------------------------------------------------

  l = rf_xxin%lower
  u = rf_xxin%upper

  IF ( (rf_xxout%lower /= l) .OR. (rf_xxout%upper /= u) ) then
    print *, "  In Shell_sort: rf_xxin and rf_xxout should"
    print *, "                 have the same shape"
    print *, "  in :", l, u
    print *, "  out:", rf_xxout%lower, rf_xxout%upper
    stop
  ENDIF

  n = u - l + 1

  CALL rf_create (a, 1, n)
  a%Val(1:n) = rf_xxin%Val(l:u)
  a%Ord(1:n) = rf_xxin%Ord(l:u)

  inc = 1
  DO WHILE (inc <= n)
    inc = 3 * inc + 1
  END DO

  DO WHILE (inc > 1)
    inc = inc / 3

    DO i = inc+1, n
      v = a%Val(i)
      p = a%Ord(i)
      j = i
      DO WHILE ((a%Val(j-inc) > v))
        a%Val(j) = a%Val(j-inc)
        a%Ord(j) = a%Ord(j-inc)
        j = j - inc
        IF (j <= inc) EXIT
      END DO

      a%Val(j) = v
      a%Ord(j) = p
    END DO
  END DO

  rf_xxout%Val(l:u) = a%Val(1:n)
  rf_xxout%Ord(l:u) = a%Ord(1:n)

  CALL rf_free (a)

  END SUBROUTINE Shell_sort

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE rf_locate (rf_xx, x, j)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!  purpose : locate index j of rf_xx%Val such that
!            rf_xx%Val(j) <= x < rf_xx%Val(j+1)
!  need : rf_xx%Val should be sorted in ASCENDING order
!  Note : if x is outside rf_xx%Val, then return jl or ju

  IMPLICIT NONE

  TYPE (RF_TYPE), intent (in) :: rf_xx
  REAL (KIND=dp), intent (in) :: x
  INTEGER, intent (out)       :: j

  INTEGER                     :: jl, ju, jm

  jl = rf_xx%lower
  ju = rf_xx%upper

  IF (x >= rf_xx%Val(ju)) THEN
    j = ju
  ELSE IF (x < rf_xx%Val(jl)) THEN
    j = jl
  ELSE

    DO WHILE ((ju-jl) > 1)
      jm = (ju + jl) / 2
      IF (x < rf_xx%Val(jm)) THEN
        ju = jm
      ELSE
        jl = jm
      END IF
    END DO
    j = jl

  ENDIF

  END SUBROUTINE rf_locate

END MODULE RF_TOOLS
